import pprint
from copy import deepcopy

from optimal_u.hex3_gen_u_optim import load_data
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import hex3_chen_old as hex3_chen
from g_process import GPModel

meas_sets = {
    1: ['T0', 'T1', 'T2', 'T3', 'Th1', 'Th2', 'Th3'],
    2: ['T0', 'Th1', 'Th2', 'Th3', 'The1', 'The2', 'The3'],
    3: ['T0', 'T', 'The1', 'The2', 'The3', 'w0', 'wh1', 'wh2', 'wh3'],
    4: ['T0', 'T', 'Th1', 'Th2', 'Th3', 'alpha1', 'alpha2']
}


def generate_meas_set_data(mset, u, d):
    global meas_sets

    d = d.to_dict('records')

    if isinstance(u, pd.DataFrame):
        u = np.array(list(zip(u['ur0'], u['ur1'])))

    results = []

    for idx, (u_, d_) in enumerate(zip(u, d)):
        # print(u_)
        # print(d_)
        result = hex3_chen.output_meas(mset, u_, d_)
        if not result['success']:
            print('Failed, small step in u')
            print(u_)

        res_dict = {}
        for idx, key in enumerate(meas_sets[mset]):
            res_dict[key] = result['y'][idx]
        results.append(res_dict)
    return np.array(results, dtype=dict)


def preprocess(u: pd.DataFrame, d: [pd.DataFrame, np.ndarray], d_order=None, u_input=None, add_noise=False):
    # Distubances
    if isinstance(d, pd.DataFrame):
        d = d.to_dict('records')

    # Optimal u values

    # This is u_chen
    if isinstance(u, pd.DataFrame):
        u_u0 = u['uc0']
        u_u1 = u['uc1']
    else:
        u_u0 = u[:, 0]
        u_u1 = u[:, 1]

    X = []

    # Ensure keys keep order
    if d_order is None:
        d_order = list(d[0].keys())
        # Remove the heat loss. Otherwise, random noise will be applied to 0 values, degrades performance.
        d_order.remove('h1')
        d_order.remove('h2')
        d_order.remove('h3')
        d_order.remove('Ts')

    # Random valve inputs
    if isinstance(u_input, pd.DataFrame):
        u_in0 = u_input['ur0']
        u_in1 = u_input['ur1']
    elif isinstance(u_input, np.ndarray):
        u_in0 = u_input[:, 0]
        u_in1 = u_input[:, 1]
    else:
        raise Exception('This should not be reached')
        u_in0 = u_u0
        u_in1 = u_u1

    d_keys = d_order.copy()
    toggle = 'alpha1' in d_keys

    if toggle:
        d_keys.remove('alpha1')
        d_keys.remove('alpha2')
    # Inputs

    np.random.seed(1)
    for idx, (u0, u1, d_row) in enumerate(zip(u_in0, u_in1, d)):
        x = [*([u0, u1] * toggle), *[d_row[k] + add_noise * np.random.normal(0, 1) for k in d_keys]]
        X.append(x)

    X = np.array(X)
    # Optimal u values
    Y = np.array(list(zip(u_u0, u_u1)))

    return X, Y, d_order


def plot_prediction(output, target, train_samples, noise, iters=0, MS=None):
    # Plotting

    fig, ax = plt.subplots(ncols=2)
    fig.suptitle(f"$u_{{opt}}$ | Iter: {iters} | "
                 f"Train samples: {train_samples} {('| MS: ' + str(MS)) * bool(MS)} | Noise: {noise}")

    ax[0].plot(output[:, 0], target[:, 0], 'b.')
    ax[0].plot([min(target[:, 0]), max(target[:, 0])], [min(target[:, 0]), max(target[:, 0])], 'k--')
    ax[0].set_xlabel('Predicted u1')
    ax[0].set_ylabel('Real u1')

    # ax[0].set_xlim([np.min(U_predicted-0.1), np.max(U_predicted)])
    # ax[0].set_ylim([np.min(U_test-0.1), np.max(U_test)])

    ax[1].plot(output[:, 1], target[:, 1], 'b.')
    ax[1].plot([min(target[:, 1]), max(target[:, 1])], [min(target[:, 1]), max(target[:, 1])], 'k--')
    ax[1].set_xlabel('Predicted u2')
    ax[1].set_ylabel('Real u2')
    # ax[1].set_xlim([np.min(U_predicted-0.1), np.max(U_predicted)])
    # ax[1].set_ylim([np.min(U_test-0.1), np.max(U_test)])
    plt.savefig(f'figs\\pred_u_{train_samples}_iter{iters}{("_MS" + str(MS)) * bool(MS)}_{noise}.eps')
    plt.close()


def get_cost(output, target, params):
    predicted_cost = np.zeros(output.shape[0])
    real_cost = np.zeros(output.shape[0])

    params_d = params['d'].to_dict('records')
    params_J = params['J_chen']
    neg_u = 0
    large_u = 0

    for idx, (u_p, u_real, param) in enumerate(zip(output, target, params_d)):
        # print(params)
        if (u_p < 0).any():
            neg_u += 1

        if (np.sum(u_p) > 1).any():
            large_u += 1
        try:
            r = hex3_chen.cost(u_p, deepcopy(param))
            cost_p = - r['J']

            assert r['success']

            cost_r = params_J.iloc[idx]['J_chen']

            assert cost_p > 0
            assert cost_r > 0
        except AssertionError:
            return predicted_cost, real_cost

        predicted_cost[idx] = cost_p
        real_cost[idx] = cost_r

    print('Negative u count', neg_u)
    print('sum U over 1:', large_u)

    return predicted_cost, real_cost


def plot_loss(predicted, target, title, fp):
    # real_cost = test_data['J_chen'].values
    plt.plot(target, predicted, 'b.')
    plt.title(title)
    plt.ylabel('Predicted cost J')
    plt.xlabel('Real cost J')
    plt.plot([min(target), max(target)], [min(target), max(target)], 'k--')
    plt.savefig(f'figs\\{fp}.eps')
    plt.close()


def plot_hist(predicted, target, title, fp):
    # histogram
    abs_value_error = np.abs(predicted - target)
    print(min(abs_value_error))
    print(max(abs_value_error))
    plt.title(title)
    plt.hist(abs_value_error, bins=100)
    plt.ylabel('Number of error points')
    plt.xlabel('Loss | Cost Function')
    plt.savefig(f'figs\\{fp}.eps')
    plt.close()


def load_train_test(train, test, meas_set=None, closed_loop=False, train_noise=False, test_noise=False, norm_y=False):
    if train_noise and test_noise:
        noise = 'TrainTest'
    elif test_noise:
        noise = 'Test'
    elif train_noise:
        noise = 'Train'  # Not tested for
    else:
        noise = 'None'

    test_data = load_data(test)
    train_data = load_data(train)

    if meas_set is not None:
        global meas_sets
        d_keys = meas_sets[meas_set]
        print('Solving measurment set data...')
        ur = train_data['u_rand']
        uc = train_data['u_chen']
        u1 = np.array(list(zip(ur['ur0'], ur['ur1'])))
        u2 = np.array(list(zip(uc['uc0'], uc['uc1'])))
        diff = u2 - u1
        split = int(0.70 * len(u1))
        u3 = np.concatenate((u1[:split] + diff[:split] * 0.98, u1[split:] + diff[split:] * 0.5))
        # u3 = u1 + diff * 0.95

        tur = test_data['u_rand']
        tuc = test_data['u_chen']
        tu1 = np.array(list(zip(tur['ur0'], tur['ur1'])))
        tu2 = np.array(list(zip(tuc['uc0'], tuc['uc1'])))
        diff = tu2 - tu1
        tu3 = tu1 + diff * 0.95
        # u = np.concatenate([u1[:-(len(u1)*40) // 100], u2[-(len(u1)*40) // 100:]])
        test_u_input = tu3

        disturbances_train = generate_meas_set_data(meas_set, u3, train_data['d'])
        disturbances_test = generate_meas_set_data(meas_set, test_u_input, test_data['d'])
    else:
        d_keys = None
        disturbances_train = train_data['d']
        disturbances_test = test_data['d']
        # Not used, dummy variable
        test_u_input = test_data['u_rand']
        u3 = train_data['u_rand']

    X_train, U_train, key_order = preprocess(train_data['u_chen'], disturbances_train, d_order=d_keys,
                                             u_input=u3, add_noise=train_noise)
    X_test, U_test, _ = preprocess(test_data['u_chen'], disturbances_test, key_order, u_input=test_u_input,
                                   add_noise=test_noise)

    # Hyperparameter
    num_restarts = 1

    mpr = GPModel()
    mpr.start()
    mpr.train(X_train, U_train, num_restarts, norm_y)
    result = mpr.test(X_test)

    U_predicted, U_covariance, m = result

    target = U_test

    plot_prediction(U_predicted, U_test, len(X_train), noise, iters=0, MS=meas_set)

    iter_u = [U_predicted]

    predicted_cost, real_cost = get_cost(U_predicted, target, test_data)

    losses = get_errors(predicted_cost, real_cost)
    pred_error = [get_errors(U_predicted, target)['RMSE']]
    RMSE_loss_iter = [losses['RMSE']]

    if closed_loop and meas_set is not None:
        i = 1
        while True:
            disturbances_test = generate_meas_set_data(meas_set, U_predicted, test_data['d'])

            X_test, _, _ = preprocess(test_data['u_chen'], disturbances_test, key_order, U_predicted,
                                      add_noise=test_noise)
            U_predicted, U_covariance, regression_info = mpr.test(X_test)

            predicted_cost, real_cost = get_cost(U_predicted, target, test_data)

            pred_error.append(get_errors(U_predicted, target)['RMSE'])
            losses = get_errors(predicted_cost, real_cost)
            RMSE_loss_iter.append(losses['RMSE'])

            iter_u.append(U_predicted)
            print(RMSE_loss_iter)
            if (np.abs(iter_u[-1] - iter_u[-2]) < 1e-8).all():
                break
            elif i == 19:
                print('Failed to converge')
                break
            i += 1

            # plot_prediction(U_predicted, U_test, len(X_train), iters=i, MS=meas_set)
        plot_prediction(U_predicted, U_test, len(X_train), noise, iters=i, MS=meas_set)

    losses = get_errors(U_predicted, target, U_covariance)

    MS = meas_set
    if MS is not None:
        plt.plot(list(range(len(iter_u))), RMSE_loss_iter)
        plt.title(f"ML Loss per iter | $u_{{opt}}$ | "
                  f"Train samples: {len(X_train)} {('| MS: ' + str(meas_set)) * bool(meas_set)} | Noise: {noise}")
        plt.ylabel('Loss - RMSE')
        plt.xlabel('Iteration')
        plt.savefig(f'figs\\iter_loss_{len(X_train)}_{("MS" + str(MS)) * bool(MS)}_{noise}.eps')
        plt.close()
    # print(*[u[0] for u in iter_u])
    # for i in [*[u[:5] for u in iter_u]]:
    #     print(i)
    # print(U_test[:5])
    # print('-' * 20)
    title = f"Loss | $u_{{opt}}$ | Train samples: {len(X_train)} " \
        f"{('| MS: ' + str(meas_set)) * bool(meas_set)} | Noise: {noise}"
    fp = f'loss_{len(X_train)}{("_MS" + str(MS)) * bool(MS)}_{noise}'
    plot_loss(predicted_cost, real_cost, title, fp)

    title = f"Histogram loss | $u_{{opt}}$ | Train samples: {len(X_train)} " \
        f"{('| MS: ' + str(meas_set)) * bool(meas_set)} | Noise: {noise}"
    fp = f'histloss_{len(X_train)}{("_MS" + str(MS)) * bool(MS)}_{noise}'
    plot_hist(predicted_cost, real_cost, title, fp)
    print('Pred max', np.max(U_predicted))
    print('Pred min', np.min(U_predicted))

    fp = f'data\\uopt_{len(X_train)}{("_MS" + str(MS)) * bool(MS)}_{noise}.csv'
    np.savetxt(fp, np.array(iter_u)[-1], delimiter=",")

    fp = f'data\\loss_{len(X_train)}{("_MS" + str(MS)) * bool(MS)}_{noise}.csv'
    np.savetxt(fp, np.array(RMSE_loss_iter), delimiter=",")

    mpr.exit()
    return losses, RMSE_loss_iter


def get_errors(predicted, target, cov=None):
    # Errors
    avg_dev = np.sum((predicted - target) ** 2) / len(target)
    avg_dev_norm = np.sum(((predicted - target) / target) ** 2) / len(target)

    errors = {'MSE': avg_dev, 'WMSE': avg_dev_norm, 'RMSE': np.sqrt(avg_dev)}

    if cov is not None:
        avg_dev_cov = np.sum((predicted - target) ** 2 / cov) / len(target)
        errors['CMSE'] = avg_dev_cov
    return errors


if __name__ == '__main__':
    samples = [500, 2500]


    # Uncomment this to show plots
    # plt.savefig = lambda x: plt.show()

    # uncommenth these two lines to save images as png
    # _savefig = plt.savefig
    # plt.savefig = lambda x: _savefig(x.replace('figs', 'imgs').replace('eps', 'png'))

    # Default saves as .eps files.

    test = '.\\datasets\\test_u_prediction2500.csv'
    import os

    if not os.path.isdir('results'):
        os.mkdir('results')

    # Black box
    pp = pprint.PrettyPrinter(indent=4)
    RESULTS = {}
    for MS in [1, 2, 3, 4]:
        RESULTS[MS] = {}
        for sample in samples:
            RESULTS[MS][sample] = {}
            train = f'.\\datasets\\train_u_prediction{sample}.csv'

            noise_label = 'No measurment noise'
            error, loss = load_train_test(train, test, meas_set=MS, closed_loop=True, train_noise=False,
                                                test_noise=False)
            RESULTS[MS][sample][noise_label] = {'errors': error, 'RMSE_Loss': loss}

            for train_noise, noise_label in zip([True, False], ['With train noise', 'without train noise']):
                error, loss = load_train_test(train, test, meas_set=MS, closed_loop=True, train_noise=train_noise,
                                                    test_noise=True)
                RESULTS[MS][sample][noise_label] = {'errors': error, 'RMSE_Loss': loss,
                                                    }
            pp.pprint(RESULTS)

    pp.pprint(RESULTS)

    train = f'.\\datasets\\train_u_prediction{500}.csv'
    error, loss = load_train_test(train, test, meas_set=None, closed_loop=False, train_noise=False,
                                        test_noise=False)
    RESULTS[0] = {}
    RESULTS[0][500] = {}

    RESULTS[0][500]['No measurment noise'] = {'errors': error, 'RMSE_Loss': loss}

    for train_noise, noise_label in zip([True, False], ['With train noise', 'without train noise']):
        error, loss = load_train_test(train, test, meas_set=None, closed_loop=False, train_noise=train_noise,
                                            test_noise=True)
        RESULTS[0][500][noise_label] = {'errors': error, 'RMSE_Loss': loss}

    pp.pprint(RESULTS)

    # # Code for comparing normalization on output.
    # RESULTS[2] = {}
    # RESULTS[2][2500] = {}
    # train = f'.\\datasets\\train_u_prediction{2500}.csv'
    # for train_noise, noise_label in zip([True, False], ['normed_y', 'not_normed_y']):
    #     error, loss, other = load_train_test(train, test, meas_set=2, closed_loop=True, train_noise=True,
    #                                   test_noise=True, norm_y=train_noise)
    #     RESULTS[2][2500][noise_label] = {'errors': error, 'MSE_Loss': loss}
    # pp.pprint(RESULTS)

